/* 
 * Copyright 1991 Oliver Jones
 * Permission to use, copy, modify, distribute, and sell this software
 * for any purpose is hereby granted without fee.  No representations
 * are made by anyone about the suitability of this software for any
 * purpose.  It is provided "as is" without express or implied
 * warranty. 
 */

#include <stdio.h>
#include <X11/Xlib.h>
#include "object.h"

/* data type declarations */
struct DrawObject {
  struct DrawObject * next;     /* link to another */
  RenderContext *     rend;     /* rendering context */
  char *              tool;     /* name of tool */
  Vector              pos[2];   /* current position */
  State               state;    /* current object state */
};

#define MIN(a,b) ((a)>(b)?(b):(a));
#define WID(a,b) ((a)>(b)?(a)-(b):(b)-(a));

#define EQ(a,b) (0 == strcmp((a),(b)))

/* vector-interface */
void EventVector (ev, v)
     XEvent * ev;
     Vector * v;
{
  switch (ev->type) {
  case ButtonPress: 
  case ButtonRelease:
    v->x = ev->xbutton.x;
    v->y = ev->xbutton.y;
    break;
  case EnterNotify: 
  case LeaveNotify:
    v->x = ev->xcrossing.x;
    v->y = ev->xcrossing.y;
    break;
  case KeyPress: 
  case KeyRelease:
    v->x = ev->xkey.x;
    v->y = ev->xkey.y;
    break;

  case MotionNotify: 
    {
      XEvent peek;
      Display * dpy = ev->xany.display;
      Window win = ev->xmotion.window;
      int x,y,x_root,y_root;
      Window child, root;
      unsigned long int mask;
      /* skip over consecutive motion events */
      while (XEventsQueued (dpy, QueuedAfterReading) &&
	     XPeekEvent(dpy, &peek),True &&
	     peek.type == MotionNotify && 
	     peek.xmotion.window==win )
	XNextEvent (dpy, ev);

      /* process motion hint if need be */
      if (ev->xmotion.is_hint) {
	XQueryPointer(ev->xany.display, ev->xany.window,
		      &root, &child, &x_root, &y_root, 
		      &x, &y, &mask);
	v->x = x;
	v->y = y;
      }
      else {
	v->x = ev->xmotion.x;
	v->y = ev->xmotion.y;
      }
    }
    break;

  default:
    v->x = 0;
    v->y = 0;
    break;
  }
}

static void makeRect (o, r)
     register DrawObject *o;
     register XRectangle *r;
{
  r->x      = MIN (o->pos[0].x, o->pos[1].x); 
  r->y      = MIN (o->pos[0].y, o->pos[1].y); 
  r->width  = WID (o->pos[0].x, o->pos[1].x); 
  r->height = WID (o->pos[0].y, o->pos[1].y);
}

/* graphic-object manipulation routines */
DrawObject * NewObject (r, tool, v1, v2)
     RenderContext *  r;
     char *           tool;
     Vector *         v1;
     Vector *         v2;
{
  DrawObject * o = (DrawObject *) XtCalloc (sizeof(DrawObject),1);
  o->next = NULL;
  o->tool = tool;
  o->rend = r;
  o->pos[0] = *v1;
  o->pos[1] = *v2;
  o->state = Erased;
  return o;
}

void PaintObject (o, newstate)
     DrawObject * o;
     State        newstate;
{
  register RenderContext * c = o->rend;
  Display * dpy = RenderDisplay (c);
  Window win = RenderWindow (c);
  XRectangle r;
  GC gc = NULL;

  switch (o->state) {

  case Erased:
    switch (newstate) {
      case Echoed:     gc = RenderEcho(c); break;
      case Drawn:      gc = RenderDraw(c); break;
    }
    break;

  case Echoed:
    switch (newstate) {
      case Unchanged:  gc = RenderEcho(c); break;
      case Erased:     gc = RenderEcho(c); break;
      case Drawn:
        PaintObject(o, Erased);
        PaintObject(o, Drawn);
        break;
      }
    break;

  case Drawn:
    switch (newstate) {
      case Unchanged:  gc = RenderDraw(c); break;
    }
    break;
  }

  if (gc) {
    if (EQ("drawrect", o->tool)) {
      makeRect (o, &r);
      XDrawRectangles (dpy, win, gc, &r, 1);
    } 
    else if (EQ("fillrect", o->tool)) {
      makeRect (o, &r);
      XFillRectangles (dpy, win, gc, &r, 1);
    }
    else if (EQ("line",o->tool)) {
      XDrawLine (dpy, win, gc, 
		 o->pos[1].x, o->pos[1].y,
		 o->pos[0].x, o->pos[0].y );
    }
    else if (EQ("arrow",o->tool)) {
      Arrow (dpy, win, gc, RenderAux(c), 3,
	     o->pos[1].x, o->pos[1].y,
	     o->pos[0].x, o->pos[0].y);
    }
  }
  if (newstate != Unchanged) o->state = newstate;
}

DrawObject * NextObject (o)
     DrawObject * o;
{
  return o->next;
}

void LinkObject (o, l)
     DrawObject * o;
     DrawObject * l;
{
  o->next = l;
}

void ResizeObject (o, v)
     DrawObject * o;
     Vector * v;
{
  State state = o->state;
  PaintObject (o, Erased);
  o->pos[1] = *v;
  PaintObject (o, state);
}

Boolean TestObject ()
{
  /* stub */
  return True;
}

void FreeObject (o)
     DrawObject * o;
{
  XtFree (o);
}

