/* 
 * Copyright 1991 Oliver Jones
 * Permission to use, copy, modify, distribute, and sell this software
 * for any purpose is hereby granted without fee.  No representations
 * are made by anyone about the suitability of this software for any
 * purpose.  It is provided "as is" without express or implied
 * warranty. 
 */

#include <stdio.h>
#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#include "render.h"
#include "image.h"
#include "object.h"

struct RenderContext {
  Display *     dpy;      /* display pointer for this object */
  Window        win;      /* window for this object */
  XColor        foreground; /* pixel value */
  XColor        background; /* pixel value */
  XColor        highlight;  /* pixel value */
  XFontStruct * font;     /* default font */
  GC            drawgc;   /* GC for drawing */
  GC            echogc;   /* GC for echoing */
  GC            auxgc;    /* extra GC */
};

static void GetColor (w, colormap, name, class, deflt, color)
     Widget w;
     Colormap colormap;
     char * name;
     char * class;
     char * deflt;
     XColor * color;
{
  XtResource res;
  String cname;

  res.resource_name = name;
  res.resource_class = class;
  res.resource_type = XtRString;
  res.resource_size = sizeof (XtRString);
  res.resource_offset = 0;
  res.default_type = XtRString;
  res.default_addr = (XtPointer) deflt;
  XtGetApplicationResources (w, &cname, res, 1, NULL, 0);
  
  XParseColor (XtDisplay(w), colormap, cname, color);
}


/* defaults from command line */
static XtResource font[] = {
  {XtNfont, XtCFont, XtRFontStruct, sizeof (XFontStruct *),
     XtOffset(RenderContext *, font), 
     XtRString, "9x15" },
};

/* create a draw context using the specified widget */
RenderContext * NewRenderContext (w)
     Widget w;
{
  Window    win = XtWindow  (w);
  Display * dpy = XtDisplay (w);
  XWindowAttributes attribs;
  Colormap colormap;
  Boolean private = False;
  RenderContext * r = 
    (RenderContext *) XtCalloc (sizeof(RenderContext),1);
 
  if (!r) return r;

  /* look up window details */
  XGetWindowAttributes (dpy, win, &attribs);
  colormap = attribs.colormap;

  switch (attribs.visual->class) {
  case GrayScale: case PseudoColor: case DirectColor:
    if (attribs.depth > 2) private = True;
    break;
  }

  XtGetApplicationResources 
    (w, r, font, XtNumber (font), NULL, 0);

  r->drawgc = XCreateGC (dpy, win, 0, NULL);
  XSetFont (dpy, r->drawgc, r->font->fid);

  r->auxgc = XCreateGC (dpy, win, 0, NULL);
  XSetFont (dpy, r->auxgc, r->font->fid);

  r->echogc = XCreateGC (dpy, win, 0, NULL);
  XSetFont (dpy, r->echogc, r->font->fid);

  GetColor (w, colormap, 
	    XtNforeground, XtCForeground, 
	    "black", &r->foreground);
  GetColor (w, colormap, 
	    XtNbackground, XtCBackground, 
	    "white", &r->background); 
  GetColor (w, colormap, 
	    "highlight", "Highlight", 
	    "gray", &r->highlight);

  if (private) {
    unsigned long int plane_mask;
    unsigned long int pixels[2];
    if (XAllocColorCells(dpy, colormap, False,
			 &plane_mask, 1, pixels, 2)) {
      /* we got the private cells.... */
      /* store colors into them */
      r->foreground.pixel=pixels[0];
      XStoreColor (dpy, colormap, &r->foreground);
      r->background.pixel=pixels[1];
      XStoreColor (dpy, colormap, &r->background);

      /* store highlight color into both overlay cells */
      r->highlight.pixel=pixels[0] | plane_mask;
      XStoreColor (dpy, colormap, &r->highlight);
      r->highlight.pixel=pixels[1] | plane_mask;
      XStoreColor (dpy, colormap, &r->highlight);

      /* set up gc for echoing */
      XSetForeground (dpy, r->echogc, plane_mask);
      XSetBackground (dpy, r->echogc, 0L);
      XSetPlaneMask (dpy, r->echogc, plane_mask);
      XSetFunction (dpy, r->echogc, GXxor);

      /* set window background */
      XSetWindowBackground(dpy, win, r->background.pixel);
      XClearArea (dpy, win, 0, 0, 0, 0, True);
    }
    else private = False;
  }

  if (!private) {
    XAllocColor (dpy, colormap, &r->foreground);
    XAllocColor (dpy, colormap, &r->background);
    XSetForeground (dpy, r->echogc,
		    r->foreground.pixel ^ r->background.pixel);
    XSetFunction (dpy, r->echogc, GXxor);
  }

  XSetForeground (dpy, r->drawgc, r->foreground.pixel);
  XSetBackground (dpy, r->drawgc, r->background.pixel);

  r->dpy = dpy;
  r->win = win;
  return r;
}

void RenderClear (r, rect, repaint)
     RenderContext * r;
     XRectangle *rect;
     Boolean repaint;
{
  XClearArea (r->dpy, r->win, 
	      rect->x, rect->y, 
	      rect->width, rect->height, 
	      repaint);
}

GC RenderDraw (r)
     RenderContext * r;
{
  return r->drawgc;
}

GC RenderEcho (r)
     RenderContext * r;
{
  return r->echogc;
}

GC RenderAux (r)
     RenderContext * r;
{
  return r->auxgc;
}

Display * RenderDisplay (r)
     RenderContext * r;
{
  return r->dpy;
}

Window RenderWindow (r)
     RenderContext * r;
{
  return r->win;
}

void RenderClip (r, region)
     RenderContext * r;
     Region region;
{
  if (region) 
    XSetRegion (r->dpy, r->drawgc, region);
  else
    XSetClipMask (r->dpy, r->drawgc, None);

  XCopyGC (r->dpy, r->drawgc, 
	   GCClipMask | GCClipXOrigin | GCClipYOrigin,
	   r->echogc);
  XCopyGC (r->dpy, r->drawgc, 
	   GCClipMask | GCClipXOrigin | GCClipYOrigin,
	   r->auxgc);
}
